/********************************************************************

  Priority depth (generalized stack distance) implementation.

  Author:  Terence Kelly  <tpkelly@eecs.umich.edu>
           http://ai.eecs.umich.edu/~tpkelly/

  Copyright 2000
  The Regents of the University of Michigan
  All rights reserved.

  Permission is granted to use, copy, create derivative works and
  redistribute this software and such derivative works for
  noncommercial education and research purposes, so long as no fee is
  charged, and so long as the copyright notice above, this grant of
  permission, and the disclaimer below appear in all copies made; and
  so long as the name of the University of Michigan is not used in
  any advertising or publicity pertaining to the use or distribution
  of this software without specific, written prior authorization.

  This software is provided as is, without representation from the
  University of Michigan as to its fitness for any purpose, and
  without warranty by the University of Michigan of any kind, either
  express or implied, including without limitation the implied
  warranties of merchantability and fitness for a particular
  purpose.  The Regents of the University of Michigan shall not be
  liable for any damages, including special, indirect, incidental, or
  consequential damages, with respect to any claim arising out of or
  in connection with the use of the software, even if it has been or
  is hereafter advised of the possibility of such damages.

********************************************************************/

static char const rcsid[] =
"$Id: stkdst.c,v 1.38 2000/02/16 05:40:54 tpkelly Exp $"
#ifdef STACK_ALGORITHM
" $Algorithm: Stack $"
#else
" $Algorithm: Splay $"
#endif
#ifndef NDEBUG
" $Debugging: Enabled $"
#else
" $Debugging: Disabled $"
#endif
;

/********************************************************************
  We assume that a reference to an item never decreases its
  priority.
********************************************************************/

#include <assert.h>
#include <errno.h>
#define __EXTENSIONS__
#include <limits.h>
#undef  __EXTENSIONS__
#include <stdio.h>
#define __EXTENSIONS__
#include <stdlib.h>
#undef  __EXTENSIONS__
#include <string.h>

#ifdef linux
#include <unistd.h>
#endif

#define Fprintf   (void)fprintf
#define Printf    (void)printf

#define INT32MAX        LONG_MAX       /* C9X-like constants */
#define INT64MAX       LLONG_MAX
#define Pd32                "ld"
#define Pd64               "lld"

typedef long int       int32_td;       /* C9X-like typedefs */
typedef long long int  int64_td;

#ifndef MAX_ITM_D           /* can override at compile time */
#define MAX_ITM_D  5252946  /* enough for March 1999 BO1 and PA */
#endif

enum { MAX_INLINE =       512,
       MAX_ITEMS  = MAX_ITM_D };

#ifdef STACK_ALGORITHM

static struct item {
  struct item *down;
  int32_td size, nrefs, lastref;
} e_items[MAX_ITEMS], e_head = { NULL, -1, INT32MAX, INT32MAX };

#define VALIDLASTREF(p) ((INT_MAX >(p)->lastref && (p)!=&e_head) || \
                         (INT_MAX==(p)->lastref && (p)==&e_head))
#define VALIDNREF(p)    (1)

#else  /* splay tree algorithm */

static struct item {
  struct item *parent, *left, *right;
  int32_td size, nrefs, lastref, rnum;
  int64_td rsum;
} e_items[MAX_ITEMS], e_root = { NULL,NULL,NULL, -1, -1, -1, 0, 0 };

#define VALIDLASTREF(p) (( 0 <= (p)->lastref && (p) != &e_root) || \
                         (-1 == (p)->lastref && (p) == &e_root))
#define VALIDNREF(p)    (( 0 <  (p)->nrefs && (p) != &e_root) || \
                         (-1 == (p)->nrefs && (p) == &e_root))

#endif

static char        e_buf[MAX_INLINE],
                  *e_progname,
                  *e_itemfile  = NULL;
static int32_td    e_inline    = 0,
                   e_num_items;
static int32_td  (*e_cmp)(const struct item *, const struct item *);

#define CMP(p,c,q) ((*e_cmp)(p, q) c 0)

#define DIE_IF(e)  do { if (e) die_if(__FILE__, __LINE__, #e); \
                      } while (0)

static void
die_if(const char * const file, const int line,
       const char * const msg) {
  Fprintf(stderr,
          "\n"
          "%s:%s:%d: fatal: `%s'\n"
          "input file `%s' line %" Pd32 ": %s\n"
          "errno == %d: `%s'\n"
          "\n"
          "usage:  %s  [-f]  -I item_file  <  reference_file\n"
          "\n"
          "-f => optional-placement Perfect LFU (default: LRU)\n"
          "\n"
          "file formats:\n"
          "   item file: size of item i on line i+1\n"
          "   reference file:  sequence of item IDs\n"
          "\n"
          "limitations:  %d chars per input line\n"
          "              items IDs in range 0..%d\n"
          "              max item size: %" Pd32 "\n"
          "\n"
          "%s\n"
          "\n",
          e_progname, file, line, msg,
          e_inline ? (e_itemfile ? e_itemfile : "(stdin)")
                   : "(none)",
          e_inline, e_inline ? e_buf : "(not reading input)",
          errno, strerror(errno), e_progname, MAX_INLINE,
          MAX_ITEMS-1, INT32MAX, rcsid);
  exit(EXIT_FAILURE);
}

#ifdef STACK_ALGORITHM

static void
set_size(int32_td ID, int32_td size) {
  assert(0 <= ID && ID < MAX_ITEMS && 0 < size);
  e_items[ID].down    = NULL;
  e_items[ID].size    = size;
  e_items[ID].nrefs   =  0;
  e_items[ID].lastref = -1;
}

static void
process(const int32_td ID, const int32_td ref,
        int32_td * const n, int64_td * const d) {
  int32_td num;
  int64_td depth;
  struct item *p, *q, *r;
  assert(0 <= ID && ID < e_num_items && 0 <= ref && n && d);
  p = &(e_items[ID]);
  assert(0 < p->size);
  p->nrefs++;                                /* update priority */
  p->lastref = ref;
  if (1 < e_items[ID].nrefs) {               /* hit */
    num = 1;
    depth = p->size;
    for (q=&e_head; q->down!=p &&
           CMP(p,<,q->down); q=q->down) {
      num++;
      depth += q->down->size;
    }
    for (r=q; r->down!=p; r=r->down) {
      num++;
      depth += r->down->size;
    }
    r->down = p->down;                       /* splice out */
  }
  else {                                     /* miss */
    for (q=&e_head; q->down &&
           CMP(p,<,q->down); q=q->down)
      /*LINTED*/
      ;
    num   = 0;
    depth = 0;
  }
  p->down = q->down;                         /* splice back in */
  q->down = p;
  *n = num;
  *d = depth;
}

#else  /* splay tree algorithm */

static void
set_size(int32_td ID, int32_td size) {
  assert(0 <= ID && ID < MAX_ITEMS && 0 < size);
  e_items[ID].parent  = NULL;
  e_items[ID].left    = NULL;
  e_items[ID].right   = NULL;
  e_items[ID].size    = size;
  e_items[ID].nrefs   = -1;
  e_items[ID].lastref = -1;
  e_items[ID].rnum    = -1;
  e_items[ID].rsum    = -1;
}

#define ROOT(p)   ((p)->parent==&e_root \
                   && e_root.right == (p))  /* is p the root?      */
#define LCHILD(p) ((p)==(p)->parent->left)  /* is p a left child?  */
#define RCHILD(p) ((p)==(p)->parent->right) /* is p a right child? */

#define IN_TREE(p)  ((p)->parent && 0 < (p)->size && 0 < (p)->nrefs \
                     && 0 <= (p)->lastref &&                        \
                     ((NULL == (p)->right && 0 == (p)->rsum) ||     \
                      ((p)->right && 0 < (p)->rsum)) &&             \
                     (LCHILD(p) || RCHILD(p)))

#define NOT_IN_TREE(p)  (NULL == (p)->parent && NULL == (p)->left  \
                         && NULL == (p)->right)

static void
tot_num_sum(const struct item * const p, int32_td * const n,
            int64_td * const d) {
  int32_td num;
  int64_td sum;
  const struct item *t, *tpar;
  assert(IN_TREE(p));
  num = 1       + p->rnum;
  sum = p->size + p->rsum;
  for (t = p; !ROOT(t); t = t->parent) {
    tpar = t->parent;
    assert(IN_TREE(tpar));
    if (LCHILD(t)) {             /* if parent is rightward, sum up */
      assert(CMP(tpar, >, p));
      num += 1 + tpar->rnum;
      sum += tpar->size + tpar->rsum;
    }
    else
      assert(CMP(tpar, <, p));
  }
  *n = num;
  *d = sum;
}

static void
rotate(struct item * const p) {
  struct item *par, *b;
  par = p->parent;
  assert(IN_TREE(p) && !ROOT(p) && IN_TREE(par));

  if (LCHILD(p)) {                      /* rotate toward right */
    b           = p->right;
    par->left   = b;
    p->right    = par;
    p->rnum    += 1         + par->rnum;
    p->rsum    += par->size + par->rsum;
  }
  else {                                /* rotate toward left */
    b           = p->left;
    par->right  = b;
    p->left     = par;
    par->rnum  -= 1       + p->rnum;
    par->rsum  -= p->size + p->rsum;
  }
  if (b)
    b->parent = par;

  p->parent = par->parent;              /* item *p goes above *par */
  if (LCHILD(par))
    p->parent->left = p;
  else
    p->parent->right = p;
  par->parent = p;
}

static void
splay(struct item * const p) {
  assert(IN_TREE(p));
  while (!ROOT(p)) {
    if (ROOT(p->parent)) {
      rotate(p);
      assert(ROOT(p));
    }
    else {
      if ((LCHILD(p) && LCHILD(p->parent)) ||   /* zig-zig */
          (RCHILD(p) && RCHILD(p->parent)))
        rotate(p->parent);
      else                                      /* zig-zag */
        rotate(p);
      rotate(p);
    }
  }
  assert(NULL == e_root.left);
}

static void
insert(struct item * const p) {
  struct item **f, *r = &e_root;
  int32_td cmp;
  assert(NOT_IN_TREE(p));
  for (;;) {
    cmp = (*e_cmp)(p, r);
    assert(0 != cmp);          /* ties are not allowed! */
    if (0 < cmp) {
      f = &(r->right);         /* p > r, go right & increment */
      r->rnum++;
      r->rsum += p->size;
    }
    else                       /* p < r, go left */
      f = &(r->left);
    if (NULL == *f) {          /* insert item */
      *f = p;
      break;
    }
    r = *f;
  }
  p->parent = r;
  splay(p);                    /* top-down splay is more efficient */
}

static int
valid_tree(const struct item *p) {
  if (p->left) {
    (void)valid_tree(p->left);
    assert(CMP(p, >, p->left));
  }
  if (p->right) {
    (void)valid_tree(p->right);
    assert(CMP(p, <, p->right));
  }
  return 1;
}

static void
delete(struct item * const p) {
  struct item *L, *R, *q;
  assert(IN_TREE(p));

  splay(p);
  /* assert(valid_tree(p)); expensive */

  if (p->left && p->right) {                     /* two children */
    L = p->left;
    R = p->right;
    for (q = L; q->right; q = q->right) /*LINTED*/
      ;
    L->parent = &(e_root);
    e_root.right = L;
    splay(q);
    /* assert(valid_tree(q));  expensive */
    assert(ROOT(q) && NULL==q->right && CMP(q,<,R));
    q->right = R;
    q->rnum = p->rnum;
    q->rsum = p->rsum;
    R->parent = q;
  }
  else {
    if (p->right || p->left) {                   /* only one child */
      q = p->left ? p->left : p->right;
      q->parent = &(e_root);
    }
    else                                         /* zero children */
      q = NULL;
    e_root.right = q;
  }

  p->parent = p->left = p->right = NULL;
  p->rnum = p->rsum = 0;

  e_root.rnum--;
  e_root.rsum -= p->size;

  assert((NULL == e_root.right && 0 == e_root.rsum) ||
         (NULL != e_root.right && 0  < e_root.rsum));
  assert(NOT_IN_TREE(p));
}

static void
process(const int32_td ID, const int32_td ref,
        int32_td * const n, int64_td * const d) {
  int32_td num;
  int64_td depth;
  struct item *p = &(e_items[ID]);
  assert(0 <= ID && ID < e_num_items && 0 <= ref);

  if (e_items[ID].parent) {          /* hit */
    tot_num_sum(p, &num, &depth);
    delete(p);
  }
  else {                             /* miss */
    num      = 0;
    depth    = 0;
    p->nrefs = 0;
    p->rsum  = 0;
    p->rnum  = 0;
  }
  p->nrefs++;                  /* update priority-related info */
  p->lastref = ref;
  insert(p);
  /* assert(valid_tree(&e_root)); expensive */
  *n = num;
  *d = depth;
}

#endif

static int32_td
lru_cmp(const struct item * const p,
        const struct item * const q) {
  assert(p && q && p != q && p->lastref != q->lastref &&
         VALIDLASTREF(p) && VALIDLASTREF(q));
  return p->lastref - q->lastref;
}

static int32_td
lfu_cmp(const struct item * const p,
        const struct item * const q) {
  assert(p && q && p != q && VALIDNREF(q) && VALIDNREF(q));
  if (p->nrefs == q->nrefs)
    return lru_cmp(p, q);
  else
    return p->nrefs - q->nrefs;
}

int
main(int argc, char *argv[]) {
  int c;
  FILE *fp;
  int32_td ID, size, num;
  int64_td depth, inf_cache_size = 0;

  e_progname = argv[0];
  e_cmp      = lru_cmp;

  DIE_IF(4 != sizeof(int32_td) || 8 != sizeof(int64_td));

  while (EOF != (c = getopt(argc, argv, "fI:")))
    switch (c) {
      case 'f':   e_cmp      = lfu_cmp;   break;
      case 'I':   e_itemfile = optarg;    break;
      default:    assert(0);              break;
    }
  DIE_IF(NULL == e_itemfile);

  /* read item size file */
  fp = fopen(e_itemfile, "r");
  DIE_IF(NULL == fp);
  for (e_inline = 1; fgets(e_buf, MAX_INLINE, fp); e_inline++) {
    DIE_IF(MAX_ITEMS < e_inline);
    size = atoi(e_buf);
    DIE_IF(0 >= size);
    set_size(e_inline-1, size);
    inf_cache_size += size;
  }
  DIE_IF(1 >= e_inline);    /* must read at least one input line */
  e_num_items = e_inline-1;
  e_itemfile  = NULL;
  e_inline    = 0;
  DIE_IF(fclose(fp));

  /* Read reference sequence. */
  for (e_inline = 1; fgets(e_buf, MAX_INLINE, stdin); e_inline++) {
    ID = atoi(e_buf);
    DIE_IF(0 > ID || ID >= e_num_items);
    process(ID, e_inline-1, &num, &depth);
    assert(0 <= depth && depth <= inf_cache_size);
    Printf("%" Pd32 " %" Pd64 " %" Pd32 "\n",
           num, depth, e_items[ID].size);
  }
  DIE_IF(1 >= e_inline);

  return 0;
}
